#
#  Type: Parsed File
#
#  This is our main data structure.
#
#  + Variables:
#    $p->{var}{NAME} = { expand => bool,                 -- 1 to expand in output, 0 to keep
#                        append => bool,                 -- 1 if defined with "+="
#                        value  => [value] }             -- value
#  + Pattern rules
#    $p->{pat}{in}{out} = { commands => [ list ],        -- command list
#                           in => in,                    -- input file type
#                           out => out }                 -- output file type
#  + Regular rules
#    $p->{rule}{out} = { commands => [ list ],           -- command list
#                        dep => [ list ] }               -- dependencies
#  + 'begin' blocks
#    $p->{sub}{NAME} = $p                                -- complete rulesets that can be instantiated
#
#  + Includes
#    $p->{includevar}[]                                  -- includes containing variables
#    $p->{includerule}[]                                 -- includes containing rules
#
#  + Files
#    $p->{files}[]
#

package ParsedFile;
use Util ('util_split', 'util_clone', 'diag_perror', 'diag_error', 'cmd_bool', 'diag_info',
          'util_format_variable', 'util_var');
use Value;
use Variable;
use Rule;
use strict;

# Create a blank Parsed File.
sub new {
    my $proto = shift;
    bless({ var => {},
            varsUsedInExpansions => {},
            pat => {},
            rule => {},
            sub => {},
            includevar => [],
            includerule => [],
            files => [] },
          ref($proto) || $proto);
}

# Given a loaded file, parse it.
sub parse {
    my $proto = shift;
    my $pLoadedFile = shift;
    my $bInstantiateVars = shift;
    my $pResult = new($proto);
    my $pRule;
    my $pData = $pLoadedFile->getData();
    my @stack;
    my $i = 0;
    my $includeTarget = 'includevar';
    my $pExpander = Expander->new($pResult);
    push @{$pResult->{files}}, $pLoadedFile;
    while ($i <= $#$pData) {
        my $me = $pData->[$i++];
        #if ($me->[0] eq 'set') {
        #    $pResult->setVar($me->[2], $me->[1], Value->new($me->[3], $pLoadedFile, $i));
        #    $includeTarget = 'includevar';
        #} elsif ($me->[0] eq 'app') {
        #    $pResult->addToVar($me->[2], $me->[1], Value->new($me->[3], $pLoadedFile, $i));
        #    $includeTarget = 'includevar';
        if ($me->[0] eq 'set' || $me->[0] eq 'app') {
            # Find variables to set
            my @vars;
            if ($bInstantiateVars && !@stack) {
                push @vars, $pExpander->expandCross($me->[2]);
            } else {
                push @vars, $me->[2];
            }
            # Attach a value if possible
            my $pValue;
            if (@vars == 1) {
                $pValue = Value->new($me->[3], $pLoadedFile, $i);
            } else {
                $pValue = Value->new($me->[3]);
            }
            # Do it
            if ($me->[0] eq 'set') {
                foreach (@vars) { $pResult->setVar($_, $me->[1], $pValue) }
            } else {
                foreach (@vars) { $pResult->addToVar($_, $me->[1], $pValue) }
            }
            $includeTarget = 'includevar';
        } elsif ($me->[0] eq 'pat') {
            $pRule = Rule->newPatternRule($me->[1], $me->[2]);
            $pResult->{pat}{$me->[1]}{$me->[2]} = $pRule;
            $includeTarget = 'includerule';
        } elsif ($me->[0] eq 'rule') {
            $pRule = $pResult->addRule($me->[1]);
            $pRule->addDepend(util_split($me->[2]));
            $includeTarget = 'includerule';
        } elsif ($me->[0] eq 'code') {
            if (!defined($pResult)) { die "internal error: code outside rule" }
            $pRule->addCommand($me->[1]);
        } elsif ($me->[0] eq 'begin') {
            push @stack, $pResult;
            #$pResult = $pResult->{sub}{$me->[1]} = ParsedFile->new();
            if (!exists $pResult->{sub}{$me->[1]}) {
                $pResult->{sub}{$me->[1]} = ParsedFile->new();
            }
            $pResult = $pResult->{sub}{$me->[1]};
        } elsif ($me->[0] eq 'end') {
            if (!@stack) { die "internal error: stack empty"; }
            $pResult = pop @stack;
        } elsif ($me->[0] eq 'include') {
            push @{$pResult->{$includeTarget}}, $me->[1];
        }
    }
    if (@stack) { die "internal error: stack not empty"; }
    # Remember variables that were used in expansions:
    foreach ($pExpander->getUsedVariables()) {
        $pResult->{varsUsedInExpansions}{$_} = 1;
    }
    $pResult;
}

# Set variable to a value. Value must be a Value object.
sub setVar {
    my ($self, $name, $expand, @value) = @_;
    my $theVariable = Variable->new($expand, 0, @value);
    $self->{var}{$name} = $theVariable;
    $theVariable;
}

# Append to a variable. If this creates the variable, remembers that
# this was an addition.
sub addToVar {
    my ($self, $name, $expand, @value) = @_;
    my $theVariable;
    if (exists($self->{var}{$name})) {
        $theVariable = $self->{var}{$name};
        $theVariable->addValue(@value);
    } else {
        $theVariable = Variable->new($expand, 1, @value);
        $self->{var}{$name} = $theVariable;
    }
    $theVariable;
}

# Append to a variable.
sub addToVarUser {
    my ($self, $name, $strVal, $sort) = @_;
    my $theVariable;
    if (!defined($sort)) { $sort = 0 }
    if (exists($self->{var}{$name})) {
        # xref Variable, Value
        $theVariable = $self->{var}{$name};
        if ($sort) {
            $theVariable->{value}[0]{value} =
              join (' ', sort util_split($theVariable->{value}[0]{value}), $strVal);
        } else {
            $theVariable->{value}[0]{value} .= " $strVal";
        }
    } else {
        $theVariable = Variable->new(0, 0, Value->new($strVal, $self->{files}[0]));
        $self->{var}{$name} = $theVariable;
    }
    $theVariable->{value}[0]->setModified();
    $theVariable;
}

# Get names of all variables.
sub getVarNames {
    keys %{$_[0]->{var}};
}

# Get variable. Returns the Variable object or undef
sub getVar {
    my ($self, $name) = @_;
    if (exists($self->{var}{$name})) { $self->{var}{$name} }
    else { undef }
}

# Get variable's value as list. This is a convenience function.
sub getVarListValue {
    my ($self, $name) = @_;
    if (exists($self->{var}{$name})) { $self->{var}{$name}->getListValue() }
    else { () }
}

# Get variable's value as string. This is a convenience function.
sub getVarStringValue {
    my ($self, $name) = @_;
    if (exists($self->{var}{$name})) { $self->{var}{$name}->getStringValue() }
    else { "" }
}

# Add a rule
sub addRule {
    my $self = shift;
    my $name = shift;
    if (exists($self->{rule}{$name})) {
        $self->{rule}{$name};
    } else {
        my $r = Rule->newRule();
        $self->{rule}{$name} = $r;
        $r->addTarget(util_split($name));
        $r;
    }
}

sub getRule {
    my $self = shift;
    my $name = shift;
    if (exists($self->{rule}{$name})) {
        $self->{rule}{$name}
    } else {
        undef
    }
}

sub clone {
    my $self = shift;
    my $copy = ParsedFile->new();
    $copy->{var} = util_clone($self->{var});
    $copy->{pat} = util_clone($self->{pat});
    $copy->{rule} = util_clone($self->{rule});
    $copy->{sub} = util_clone($self->{sub});
    $copy->{includevar} = util_clone($self->{includevar});
    $copy->{includerule} = util_clone($self->{includerule});
    $copy->{files} = util_clone($self->{files});
    $copy->{varsUsedInExpansions} = util_clone($self->{varsUsedInExpansions});
    $copy;
}

# Merge another Parsed File into this one. Note that this is *not* a
# symmetric operation. It modifies this object. It assumes that the
# object we're merging in is a temporary, and potentially shares/modifies
# its parts.
sub merge {
    my $self = shift;
    my $other = shift;

    # Files and Includes:
    push @{$self->{files}}, @{$other->{files}};
    push @{$self->{includerule}}, @{$other->{includerule}};
    push @{$self->{includevar}}, @{$other->{includevar}};

    # Variables:
    foreach (keys %{$other->{var}}) {
        if (exists ($self->{var}{$_})) {
            $self->{var}{$_} =
              $self->{var}{$_}->merge($other->{var}{$_});
        } else {
            $self->{var}{$_} = $other->{var}{$_};
        }
    }

    # Rules:
    foreach (keys %{$other->{rule}}) {
        my $newrule = $self->addRule($_);
        # FIXME: xref Rule, or add a merge()
        $newrule->addDepend(@{$other->{rule}{$_}{depend}});
        $newrule->addCommand(@{$other->{rule}{$_}{commands}});
    }

    # Pattern Rules:
    foreach my $in (keys %{$other->{pat}}) {
        # FIXME: this means pattern rules defined in Settings
        # override those from Rules-xx.mak
        foreach my $out (keys %{$other->{pat}{$in}}) {
            if (!exists $self->{pat}{$in}{$out}) {
                $self->{pat}{$in}{$out} = $other->{pat}{$in}{$out};#->clone();
            }
        }
    }

    # Blocks
    foreach (keys %{$other->{sub}}) {
        if (exists $self->{sub}{$_}) {
            $self->{sub}{$_}->merge($other->{sub}{$_});
        } else {
            $self->{sub}{$_} = $other->{sub}{$_};#->clone();
        }
    }

    # Used variables:
    foreach (keys %{$other->{varsUsedInExpansions}}) {
        $self->{varsUsedInExpansions}{$_} = 1;
    }

    $self;
}

# Display this file on stdout in parsed form. With cmd_bool(verbose) set,
# make more detailed output.
sub dump {
    my $self = shift;
    my $prefix = shift;
    print $prefix, "Files:\n";
    foreach (@{$self->{files}}) {
        print "$prefix  ", $_->getName(), "\n";
    }
    print $prefix, "Variables:\n";
    foreach (sort keys %{$self->{var}}) {
        my $var = $self->{var}{$_};
        my $value = $var->getStringValue();
        if ($var->{expand}) {
            print "$prefix  $_ := $value\n";
        } else {
            print "$prefix  $_ = $value\n";
        }
        # xref Variable:
        if (cmd_bool('verbose')) {
            foreach (@{$var->{value}}) {
                if (defined($_->getFile())) {
                    print "$prefix     # ", $_->getFile()->getName(), ":", $_->getLine(), ": ", $_->getValue(), "\n";
                }
            }
            print "$prefix     # as list:\n";
            foreach ($var->getListValue()) {
                print "$prefix     #   $_\n";
            }
        }
    }
    print $prefix, "Rules:\n";
    foreach (sort keys %{$self->{rule}}) {
        # xref Rule
        print "$prefix  $_ : ", join(' ', @{$self->{rule}{$_}{depend}}), "\n";
        foreach (@{$self->{rule}{$_}{commands}}) {
            print "$prefix      $_\n";
        }
    }
    print $prefix, "Pattern Rules:\n";
    foreach my $in (sort keys %{$self->{pat}}) {
        foreach my $out (sort keys %{$self->{pat}{$in}}) {
            print "$prefix  .$in.$out\n";
            # xref Rule:
            if (cmd_bool('verbose') && $self->{pat}{$in}{$out}{used}) {
                print "$prefix        # this rule is used\n";
            }
            foreach (@{$self->{pat}{$in}{$out}{commands}}) {
                print "$prefix        $_\n";
            }
        }
    }
    print $prefix, "Templates:\n";
    foreach (sort keys %{$self->{sub}}) {
        print "$prefix  $_:\n";
        $self->{sub}{$_}->dump("$prefix    ");
    }

    if (cmd_bool('verbose')) {
        print $prefix, "Namespace Candidates:\n";
        my $p = $self->findNamespaceCandidates();
        foreach (sort keys %$p) {
            print "$prefix  $_\n";
        }
    }
}

# Given a list of files, see what we can produce from them. This looks
# at regular rules as well as pattern rules. Pattern rules will be
# instantiated appropriately.
sub processFiles {
    my $self = shift;
    my $pFiles = shift;

    my $pNsCands = $self->findNamespaceCandidates();

    # We want to do the following:
    # - for each input file, generate a set of ALL output files
    #   as $ruleOutput{in}{out} = N, where N is
    #     0 = this is the identity transformation
    #     1 = application of a regular rule
    #     2 = application of a pattern rule
    # - use pattern rules as well as regular rules

    my %ruleOutput;

    # First pass: identity transformations
    foreach (@$pFiles) {
        $ruleOutput{$_}{$_} = 0;
    }

    # Second pass: regular rules
    # xref Variable
    foreach my $thisRuleOutput (keys %{$self->{rule}}) {
        my $thisRule = $self->{rule}{$thisRuleOutput};
        foreach my $thisRuleInput (@{$thisRule->{depend}}) {
            foreach my $o (util_split($thisRuleOutput)) {
                $ruleOutput{$thisRuleInput}{$o} = 1;
            }
        }
    }

    # Third pass: process pattern rules
    my @todo = sort @$pFiles;
    my %done;

    my %allStems;
    my $renameFlag = Util::util_bool($self->getVarStringValue('PROJ_UNIQUE_STEMS'));

    foreach (@todo) { $done{$_} = 1 }
    while (@todo) {
        my $thisFile = shift @todo;
        if ($thisFile =~ m|^(.*)\.([^/.]+)$|) {
            my $stem = $1;
            my $ext = $2;
            foreach my $outExt (keys %{$self->{pat}{$ext}}) {
                my $outFile = "$stem.$outExt";
                my $thisRule = $self->{pat}{$ext}{$outExt};
                if (!exists $done{$outFile}) {
                    # Rename it?
                    if ($renameFlag && !defined($self->getVar('*_'.$thisFile))) {
                        # We want to rename files so that they have unique stems (basenames),
                        # because archive files cannot reliably distinguish files in directories.
                        # One problem with this auto-rename is that it can change the renaming
                        # of one file when another file is added. Users must rebuild libraries
                        # when this happens.
                        my ($dir, $base);
                        if ($stem =~ m|^(.*/)(.*)|) {
                            $dir = $1; $base = $2;
                        } else {
                            $dir = ''; $base = $stem;
                        }
                        my $n = 0;
                        my $new = $base;
                        while (exists $allStems{$new} && $allStems{$new} ne $stem) {
                            # diag_info("Conflict with '$allStems{$new}' and '$stem'");
                            $new = $base.++$n;
                        }
                        $allStems{$new} = $stem;
                        if ($n) {
                            # We have to rename it
                            diag_info("generate: renaming '$thisFile' -> '$dir$new'");
                            $self->setVar('*_'.$thisFile, 1, Value->new($dir.$new));

                            # Rename it for output
                            $outFile = "$dir$new.$outExt";

                            # Trigger pattern rule instantiation
                            $pNsCands->{$thisFile} = 1;
                        }
                    }

                    # File not (scheduled for) processing
                    $done{$outFile} = 1;
                    $thisRule->markUsed();
                    push @todo, $outFile;
                    if (exists ($pNsCands->{$thisFile})) {
                        # This may rename the outfile!
                        $outFile = $self->instantiatePatternRule($thisRule, $stem);
                    }
                    $ruleOutput{$thisFile}{$outFile} = 2; # FIXME --?
                }
            }
        }
    }

    # Fourth pass: build closure
    my $pass = 0;
    while (1) {
        my $did = 0;
        foreach my $inputFile (keys %ruleOutput) {
            foreach my $intermediateFile (keys %{$ruleOutput{$inputFile}}) {
                if (exists $ruleOutput{$intermediateFile}) {
                    foreach my $outputFile (keys %{$ruleOutput{$intermediateFile}}) {
                        if (!exists $ruleOutput{$inputFile}{$outputFile}) {
                            $ruleOutput{$inputFile}{$outputFile} = 3;
                            ++$did;
                        }
                    }
                }
            }
        }
        last if !$did;
        ++$pass;
        diag_info("generate: pass $pass yielded $did files that can be built using a rule combination");
    }

    \%ruleOutput;
}

# Get set of all possible candidates for namespaced variable access.
# This is used to tell in advance whether a pattern rule instantiation
# has a chance to be useful, and saves a costly instantiation.
sub findNamespaceCandidates {
    my $self = shift;
    my %result;
    foreach (keys %{$self->{var}}) {
        my @list = split /_/, $_;
        if (@list > 1) {
            shift @list;
            while (@list) {
                $result{join('_', @list)} = 1;
                shift @list;
            }
        }
    }
    \%result;
}

# (Try to) instantiate a pattern rule
# Parameters:
#   - the pattern rule
#   - the stem
sub instantiatePatternRule {
    my $self = shift;
    my $pat = shift;
    my $stem = shift;
    my $insuffix = $pat->{in};
    my $outsuffix = $pat->{out};

    my $outfile = $stem . '.' . $outsuffix; # xref Rule

    # We instantiate $*, $< and $@, which are the variables commonly used
    # in pattern rules. Although common Makes provide some or even all of them
    # in an explicit rule, DOS/Windows Makes do not or even provide contradictory
    # definitions, such as making both $* and $< the target file (Borland).
    # We define $@ in terms of $* to allow users to rename output files.
    my $expander = Expander->new($self);
    $expander->addLocalVariable('*', $stem);
    $expander->addLocalVariable('<', $stem.'.'.$insuffix);
    $expander->addLocalVariable('@', $expander->expandVars('$*.'.$outsuffix));
    if ($stem =~ m{(.*)/(.*)}) {
        $expander->addLocalVariable('*D', $1);
        $expander->addLocalVariable('<D', $1);
        $expander->addLocalVariable('@D', $1);
        $expander->addLocalVariable('*F', $2);
        $expander->addLocalVariable('<F', $2.'.'.$insuffix);
        $expander->addLocalVariable('@F', $2.'.'.$outsuffix);
    } else {
        $expander->addLocalVariable('*D', '.');
        $expander->addLocalVariable('<D', '.');
        $expander->addLocalVariable('@D', '.');
        $expander->addLocalVariable('*F', $stem);
        $expander->addLocalVariable('<F', $stem.'.'.$insuffix);
        $expander->addLocalVariable('@F', $stem.'.'.$outsuffix);
    }

    # First expand the rules without special namespace:
    my @normal;
    foreach (@{$pat->{commands}}) { # xref Rule
        push @normal, $expander->expandRefs($_);
    }

    # Now do the same with this specific target: xref Rule
    my @specific;
    $expander->addNamespace("_" . $stem . "." . $insuffix);
    $expander->addLocalVariable('@', $expander->expandVars('$*.'.$outsuffix));
    foreach (@{$pat->{commands}}) {
        push @specific, $expander->expandRefs($_);
    }

    # User might have renamed $@
    my $newOutfile = $expander->expandVars('$@');

    # Compare
    my $ok = ($outfile eq $newOutfile);
    foreach (0 .. $#normal) {
        if ($normal[$_] ne $specific[$_]) { $ok = 0; last }
    }

    # If they are unequal, instantiate it
    if (!$ok) {
        # xref Rule
        # FIXME: it would make sense to make the input file first
        my $r = $self->addRule($newOutfile);
        $r->addDepend($stem . '.' . $pat->{in});
        $r->addCommand(@specific);
    }
    $newOutfile;
}

# Instantiate a file (sub block) into this ParsedFile using an
# Expander. Does not do parent processing.
sub instantiateFile {
    my $self = shift;
    my $other = shift;
    my $expander = shift;

    # Variables:
    foreach (keys %{$other->{var}}) {
        # Expand name and value
        my $newname = $expander->expandVars($_);
        my $value   = $expander->expandRefs($other->{var}{$_}->getStringValue());
        my $newvar  = Variable->new($other->{var}{$_}->isExpand(),
                                    $other->{var}{$_}->isAppend(),
                                    Value->new($value));

        if (exists($self->{var}{$newname})) {
            $self->{var}{$newname} = $self->{var}{$newname}->merge($newvar)
        } else {
            $self->{var}{$newname} = $newvar;
        }
    }

    # Rules:
    foreach (keys %{$other->{rule}}) {
        my $newname = $expander->expandRefs($_);
        my $newrule = $self->addRule($newname);
        foreach (@{$other->{rule}{$_}{depend}}) {
            my $str = $expander->expandRefs($_);
            $newrule->addDepend($str)
              if $str ne '';
        }
        foreach (@{$other->{rule}{$_}{commands}}) {
            my $str = $expander->expandRefs($_);
            $newrule->addCommand($str)
              if $str ne '';
        }
    }

    # I don't think we should instantiate pattern rules. If we do, we should
    # limit them to this instantiation's files, i.e. instantiate them for
    # everything in $(FILES). But that would counter the idea that
    # everything done to a file solely depends on its extension.

    # Pattern Rules:
    # foreach (keys %{$other->{pat}}) {
    #     $self->{pat}{$_} = clone($other->{pat}{$_});    # FIXME?
    # }
}

# Instantiate a file (sub block)'s assignments.
sub instantiateAssignments {
    my $self = shift;
    my $other = shift;
    my $expander = shift;

    # Variables:
    foreach (keys %{$other->{var}}) {
        # Expand name and value
        if (!$other->{var}{$_}->isAppend()) {
            my $newname = $expander->expandVars($_);
            my $value   = $expander->expandRefs($other->{var}{$_}->getStringValue());
            my $newvar  = Variable->new($other->{var}{$_}->isExpand(),
                                        $other->{var}{$_}->isAppend(),
                                        Value->new($value));
            $self->{var}{$newname} = $newvar;
        }
    }
}

# Get list of all suffixes marked as used
sub getUsedSuffixes {
    my $self = shift;
    my $pPatHash = $self->{pat};
    my %resultSet;
    foreach my $input (sort keys %$pPatHash) {
        foreach my $output (sort keys %{$pPatHash->{$input}}) {
            if ($pPatHash->{$input}{$output}{used}) {
                $resultSet{$input} = 1;
                $resultSet{$output} = 1;
            }
        }
    }
    sort keys %resultSet;
}

# Perform all variable updates. Returns list of all files that need to
# be updated.
sub performUpdates {
    my $self = shift;

    my @return;

    foreach my $varName (keys %{$self->{var}}) {
        my $var = $self->{var}{$varName};
        # xref Variable
        foreach my $value (@{$var->{value}}) {
            if ($value->isModified()) {
                my $file = $value->getFile();
                if (defined($file)) {
                    # Mark the file
                    if (!grep{$_ == $file} @return) { push @return, $file }

                    # Generate new text
                    my $op;
                    if ($var->isAppend()) {
                        $op = "+=";
                    } elsif ($var->isExpand()) {
                        $op += ":="
                    } else {
                        $op = "=";
                    }

                    # FIXME: handle blank lines?
                    my $text = util_format_variable($varName, $op, $value->getValue());

                    # Place it in file
                    my $line = $value->getLine();
                    if (defined($line)) {
                        $file->setLine($line, ['raw', $text]);
                    } else {
                        $file->addLine(['raw', $text]);
                    }
                }
            }
        }
    }

    \@return;
}

# Check whether we need to reload the project. This is the case when a variable
# was changed that was used on the LHS of an assignment, as in
#   CFLAGS_$(FILES) = -O
# and thus produces cross-product semantics in variable generation.
sub isReloadRequired {
    my $self = shift;

    foreach my $varName (keys %{$self->{var}}) {
        my $var = $self->{var}{$varName};
        # xref Variable
        foreach my $value (@{$var->{value}}) {
            if ($value->isModified()) {
                # Variable is modified
                if (exists $self->{varsUsedInExpansions}{$varName}) {
                    return 1;
                }
            }
        }
    }
    return 0;
}

# Rename variables to safe names. We can handle almost any non-whitespace
# as variable name, but not all Makes can. Here, we rename them to "safe"
# names. Instead of rewriting each and every item in the project, we add
# new expandable variables which are replaced by the new variable name by
# the output engine.
sub safeRenameVariables {
    my $self = shift;
    my $pVars = $self->{var};
    foreach (sort keys %$pVars) {
        my $var = $pVars->{$_};
        if (/[^A-Za-z_0-9]/ && !$var->isExpand()) {
            # We want to rename this variable, so figure out a new name
            my $newName = $_;
            $newName =~ s/[^A-Za-z_0-9]/_/g;
            $newName =~ s/__+/_/g;
            if (exists $pVars->{$newName}) {
                my $suffix = 0;
                while (exists $pVars->{$newName."_".$suffix}) { ++$suffix }
                $newName .= "_".$suffix;
            }
            diag_info("generate: renaming '$_' to '$newName'");

            # Rename it by moving to new name
            $pVars->{$newName} = $pVars->{$_};

            # Replace original variable by a dummy
            $pVars->{$_} = Variable->new(1, 0, Value->new(util_var($newName)));

            # FIXME: this does not preserve the variable's isAppend field.
            # However, that field is no longer relevant during generation.
        }
    }
}

1;
